/*
 * dsm.h - headers for Unihedron DSM interface
 *
 * $Id: dsm.h,v 1.12 2008/08/19 03:13:17 johns Exp $
 */

#ifdef  __cplusplus
extern "C" {
#endif

#define DSM_ERR_NONE             0
#define DSM_ERR_OPEN            -1
#define DSM_ERR_MISC            -2
#define DSM_ERR_BADHANDLE       -3
#define DSM_ERR_UNSUPPORTED     -4
#define DSM_ERR_BADALLOC        -5
#define DSM_ERR_BADPARAM        -6
#define DSM_ERR_COMM_TIMEOUT    -7

/* bitwise ORed flags specifying what data we've received so far */
#define DSM_COMM    0x01  /* we've received some valid DSM packets   */
#define DSM_DATA    0x02  /* DSM has returned valid integrations */
#define DSM_SIDEA   0x04  /* Side A data is recorded */
#define DSM_SIDEB   0x08  /* Side B data is recorded */
#define DSM_TEMP    0x10  /* Temp data is recorded   */
#define DSM_PERIOD  0x20  /* Integration period is recorded */
#define DSM_TEST    0x40  /* A/D TEST mode, not real exposure ADUs */

/* constants for aperture wheel positioning */
#define DSM_APERTURE_IRFILTER 0   /* simulate visual response        */
#define DSM_APERTURE_BLACK    5   /* for taking darks                */
#define DSM_APERTURE_OPEN     10  /* full spectrum                   */
#define DSM_APERTURE_WHITE    15  /* reflect daytime sun from sensor */

typedef void * DSMHandle; /* DSM Handle type, used by all dsm API functions */

/*
 * dsm_open()
 *   Open a named serial port which a DSM is attached to.
 * Returns a handle which is used by all other dsm API functions.
 * If the serial port open fails, or the dsm does not pass initialization
 * tests, then a NULL is returned as the handle.
 */
DSMHandle dsm_open(const char * dsmname, int highspeed);

/*
 * dsm_close()
 *   Closes down the DSM serial port, frees allocated resources and
 * discards any unprocessed dsm messages.
 */
int dsm_close(DSMHandle voidhandle);


/*
 * dsm_flushbuffer()
 *   Flush all pending input characters from the buffer
 */
int dsm_flushbuffer(DSMHandle voidhandle);


/*
 * dsm_set_test_mode()
 *   Set the DSM to test mode or normal mode.
 */
int dsm_set_test_mode(DSMHandle voidhandle, int on);

/*
 * dsm_reset()
 *   Reset the DSM unit.  A red LED can be seen through the 
 * condenser lens during reset.
 */
int dsm_reset(DSMHandle voidhandle);

/* 
 * dsm_set_integration_count()
 *   Sets the length of time to integrate samples, where 
 *   each count is 102.4us.  The minimum allowable count parameter
 *   is 10, and the maximum is 9765.
 */
int dsm_set_integration_count(DSMHandle voidhandle, int count);

/* 
 * dsm_set_integration_time()
 *   Sets the length of time (in seconds) to integrate samples
 */
int dsm_set_integration_time(DSMHandle voidhandle, float time);

/* 
 * dsm_set_aperture()
 *   Command the aperture wheel to move to the requested position
 */
int dsm_set_aperture(DSMHandle voidhandle, int aperture);

/*
 * dsm_clear_status()
 *   Clear status before subsequent update calls
 */
int dsm_clear_status(DSMHandle voidhandle);

/*
 * dsm_syncup_status()
 *   Read the DSM serial port until no packets are still pending
 */
int dsm_syncup_status(DSMHandle voidhandle);

/*
 * dsm_get_status()
 *   Polls the DSM serial port for new packets
 */
int dsm_get_status(DSMHandle voidhandle);


/*
 * dsm_get_soonest_statuspair()
 *   Polls the DSM serial port for new packets, waiting until readings are
 * available from both sides of the A/D converter.
 */
int dsm_get_soonest_statuspair(DSMHandle voidhandle);

/*
 * dsm_get_unique_statuspair()
 *   Polls the DSM serial port for new packets, waiting until a new pair 
 * of unique (not previously sampled) readings are
 * available from both sides of the A/D converter.
 */
int dsm_get_unique_statuspair(DSMHandle voidhandle);

/*
 * dsm_get_unique_array()
 *   Polls the DSM serial port for new packets, waiting until the requested
 * number of unique (not previously sampled) readings are
 * available from both sides of the A/D converter.
 */
int dsm_get_unique_statusarray(DSMHandle voidhandle, int numsamples,
                               int *range, int *sidea, int *sideb);

/*
 * dsm_print_aperture_status()
 *   Print collected data to stdout
 */
int dsm_print_aperture_status(DSMHandle voidhandle);

/*
 * dsm_print_status()
 *   Print collected data to stdout
 */
int dsm_print_status(DSMHandle voidhandle);


#ifdef  __cplusplus
}
#endif



