/*
 * gps.c -- A simple driver for NMEA-0183 GPS   
 *
 *  Copyright 1999 John E. Stone (j.stone@acm.org)
 *
 *  $Id: gps.c,v 1.1 2007/09/02 06:42:04 johns Exp $
 *
 */

#define _POSIX_SOURCE 1

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

#include "gps.h"
#include "serial.h"

#define SEQRMC ('R' | ('M' << 8) | ('C' << 16))
#define SEQGLL ('G' | ('L' << 8) | ('L' << 16))
#define SEQGGA ('G' | ('G' << 8) | ('A' << 16))

typedef struct {
  RS232CommHandle commhandle;
  char buf[1024];
  int bufpos;
  int packfinished;
  int statusflags;    /* which data is valid */
  int serialno;       /* serial number of received data */
  char talkerid[3];   /* NMEA talker ID   */
  char seqid[4];      /* NMEA sequence ID */
  char cksum;         /* checksum of received data */
  float longtitude;   /* longtitude */
  char longdir;       /* longtitude direction */
  float latitude;     /* latitude */
  char latdir;        /* latitude direction */
  int fixquality;     /* GPS fix quality */
  int numsats;        /* number of satellites being tracked */ 
  float dop;          /* dilution of position */ 
  int day;            /* date */
  int month;
  int year;
  int hours;          /* time of packet, hours */
  int mins;           /* time of packet, minutes */
  int secs;           /* time of packet, seconds */
} gpshandle;


static char * getfield(char * str) {
  int i=0;
  char * end;

  while ((str[i] != '\0') && (str[i] != ',') && (str[i] != '*')) {
    i++;
  }
  str[i] = '\0';
  end = &str[i+1];
  return end;
}

int gps_flushbuffer(GPSHandle voidhandle) {
  gpshandle * handle = (gpshandle *) voidhandle;
  while (rs232_comm_read(handle->commhandle, handle->buf, 256) > 0);
  return 0;
}

GPSHandle gps_open(const char * gpsname) {
  gpshandle * handle;

  if (gpsname == NULL)
    return NULL;

  handle = (gpshandle *) malloc(sizeof(gpshandle));
  if (handle == NULL)
    return NULL;

  /* clear all values in gpshandle to 0 */
  memset(handle, 0, sizeof(gpshandle));
  handle->bufpos = 0;   
  handle->packfinished = 0;
  handle->serialno = 0;
  handle->statusflags = 0;

  if (rs232_comm_open(gpsname, &handle->commhandle, 4800) == -1) {
    free(handle);
    return NULL;
  }

  gps_flushbuffer(handle); /* flush any existing incoming data */

  return handle; /* successfull open */
}


int gps_close(GPSHandle voidhandle) {
  gpshandle * handle = (gpshandle *) voidhandle;

  if (handle == NULL)
    return -1;

  rs232_comm_close(&handle->commhandle);
  free(handle); 
  return 0; /* successfull close */
}

int gps_clear_data(GPSHandle voidhandle) {
  gpshandle * handle = (gpshandle *) voidhandle;
  handle->statusflags = 0;
  return 0;
}

int gps_have_data(GPSHandle voidhandle, int *havedata) {
  gpshandle * handle = (gpshandle *) voidhandle;
  *havedata = handle->statusflags;
  return 0;
}

static int gps_update(GPSHandle voidhandle) {
  int i, j, num, packs;
  unsigned char rawbuf[1024];
  gpshandle * handle = (gpshandle *) voidhandle;
  int seqcode;  

  packs = 0;

  if (handle == NULL)
    return -1;

 
  packs = 0; /* no packs received yet */
  num = rs232_comm_read(handle->commhandle, (char *) rawbuf, 1023);
 
  if (num > 0) {
    for (i=0; i<num; i++) {
      /* start of new packet */
      if (rawbuf[i] == '$') {
        handle->bufpos = 0;   /* skip bytes */
        handle->packfinished = 0;
        continue;
      }

      handle->buf[handle->bufpos++] = rawbuf[i];

      /* checksumming */
      if (rawbuf[i] == '*') {
        handle->cksum = 0;
        for (j=0; j<handle->bufpos - 1; j++) {
          handle->cksum ^= handle->buf[j]; 
        }
      }

      /* end of packet */
      if (rawbuf[i] == '\n') {
        char dcksum[3];
        int cksum;
        handle->packfinished = 1; /* last char of packet */
        handle->buf[handle->bufpos++] = '\0'; 
      
        /* checksum test */ 
        dcksum[0] = handle->buf[handle->bufpos - 5];
        dcksum[1] = handle->buf[handle->bufpos - 4];
        dcksum[3] = '\0';
        sscanf(dcksum, "%02x", &cksum);

        if (cksum != handle->cksum) {
#if 0
          printf("GPS Checksum error!\n");
          printf("  Computed: %02x  Transmitted: %02x\n", handle->cksum, cksum);
#endif
          continue;
        }
      }

      if (!handle->packfinished)
        continue;

      /* Handle proprietary data packets */
      if (handle->buf[0] == 'P') {
        /* printf("Proprietary. \n"); */

        /* reset */ 
        handle->bufpos = 0;   
        handle->packfinished = 0;
        packs++;

        continue;
      }

      handle->talkerid[0] = handle->buf[0];
      handle->talkerid[1] = handle->buf[1];
      handle->talkerid[2] = '\0';

      handle->seqid[0] = handle->buf[2];
      handle->seqid[1] = handle->buf[3];
      handle->seqid[2] = handle->buf[4];
      handle->seqid[3] = '\0';

      seqcode =  handle->seqid[0] | 
                (handle->seqid[1] << 8) | 
                (handle->seqid[2] << 16); 

#if 0
      printf("Talker: %s  Seq: %s\n", handle->talkerid, handle->seqid);
#endif
      switch (seqcode) {
        case SEQRMC: 
          { char blah[1024];
            char * next, * cur;
            cur = (char *) handle->buf;

            /* eat the first field */
            next = getfield(cur);
            cur=next;

            next = getfield(cur); /* POS_UTC */
            cur=next;
            next = getfield(cur); /* POS_STAT */
            cur=next;
            next = getfield(cur); /* LAT */
            cur=next;
            next = getfield(cur); /* LAT_REF */
            cur=next;
            next = getfield(cur); /* LON */
            cur=next;
            next = getfield(cur); /* LON_REF */
            cur=next;
            next = getfield(cur); /* SPD */
            cur=next;
            next = getfield(cur); /* HDG */
            cur=next;

            next = getfield(cur); /* DATE */
            if (sscanf(cur, "%s", blah) > 0) {
              int day, month, year;
    
              if (sscanf(&blah[4], "%02d", &year) == 1) {
                blah[4]='\0';
                if (sscanf(&blah[2], "%02d", &month) == 1) { 
                  blah[2]='\0';
                  if (sscanf(blah, "%02d", &day) == 1) {
                    handle->day   = day;
                    handle->month = month;
                    handle->year  = year + 2000;
                    handle->statusflags |= GPS_DATE;
                  }
                }
              }
            }
            cur=next;

            next = getfield(cur); /* MAG_VAR */
            cur=next;
            next = getfield(cur); /* MAG_REF */
            cur=next;
          }
          break;


        case SEQGGA: 
          { char blah[1024];
            char * next, * cur;
            cur = (char *) handle->buf;

            /* eat the first field */
            next = getfield(cur);
            cur=next;

            next = getfield(cur);
            if (sscanf(cur, "%s", blah) > 0) {
              int hour, min, sec;
    
              if (sscanf(&blah[4], "%02d", &sec) == 1) {
                blah[4]='\0';
                if (sscanf(&blah[2], "%02d", &min) == 1) { 
                  blah[2]='\0';
                  if (sscanf(blah, "%02d", &hour) == 1) {
                    handle->hours = hour;
                    handle->mins  = min;
                    handle->secs  = sec;
                    handle->statusflags |= GPS_TIME;
                  }
                }
              }
            }
            cur=next;

            next = getfield(cur);
            if (sscanf(cur, "%s", blah) > 0) {
              int deg;
              float min;

              if (sscanf(&blah[2], "%f", &min) == 1) {
                blah[2]='\0';
                if (sscanf(blah, "%d", &deg) == 1) {
                  handle->latitude = deg + (min / 60.0f);
                  handle->statusflags |= GPS_FIXLAT;
                }
              }
            }
            cur=next;

            next = getfield(cur);
            sscanf(cur, "%c", &handle->latdir); 
            cur=next;

            next = getfield(cur);
            if (sscanf(cur, "%s", blah) > 0) {
              int deg;
              float min;
    
              if (sscanf(&blah[3], "%f", &min) == 1) {
                blah[3]='\0';
                if (sscanf(blah, "%d", &deg) == 1) {
                  handle->longtitude = deg + (min / 60.0f); 
                  handle->statusflags |= GPS_FIXLONG;
                }
              }
            }
            cur=next;

            next = getfield(cur);
            sscanf(cur, "%c", &handle->longdir);
            cur=next;

            next = getfield(cur);
            if (sscanf(cur, "%d", &handle->fixquality) == 1)
              handle->statusflags |= GPS_FIXQ;
            cur=next;

            next = getfield(cur);
            if (sscanf(cur, "%02d", &handle->numsats) == 1)
              handle->statusflags |= GPS_SATS;
            cur=next;

            next = getfield(cur);
            if (sscanf(cur, "%f", &handle->dop) == 1)
              handle->statusflags |= GPS_DOP;
            cur=next;

            /* eat the next field */
            next = getfield(cur);
            cur=next;

            /* eat the next field */
            next = getfield(cur);
            cur=next;

            /* eat the next field */
            next = getfield(cur);
            cur=next;

            /* eat the next field */
            next = getfield(cur);
            cur=next;
          }
          break;

        case SEQGLL: 
          { char blah[1024];
            char * next, * cur;
            cur = (char *) handle->buf;

            /* eat the first field */
            next = getfield(cur);
            cur=next;

            next = getfield(cur);
            if (sscanf(cur, "%s", blah) > 0) {
              int deg;
              float min;
    
              if (sscanf(&blah[2], "%f", &min) == 1) {
                blah[2]='\0';
                if (sscanf(blah, "%d", &deg) == 1) {
                  handle->latitude = deg + (min / 60.0f);
                  handle->statusflags |= GPS_FIXLAT;
                }
              }
            }
            cur=next;

            next = getfield(cur);
            sscanf(cur, "%c", &handle->latdir); 
            cur=next;

            next = getfield(cur);
            if (sscanf(cur, "%s", blah) > 0) {
              int deg;
              float min;
    
              if (sscanf(&blah[3], "%f", &min) == 1) {
                blah[3]='\0';
                if (sscanf(blah, "%d", &deg) == 1) {
                  handle->longtitude = deg + (min / 60.0f); 
                  handle->statusflags |= GPS_FIXLONG;
                }
              }
            }
            cur=next;

            next = getfield(cur);
            sscanf(cur, "%c", &handle->longdir);
            cur=next;

            next = getfield(cur);
            if (sscanf(cur, "%s", blah) > 0) {
              int hour, min, sec;
    
              if (sscanf(&blah[4], "%02d", &sec) == 1) {
                blah[4]='\0';
                if (sscanf(&blah[2], "%02d", &min) == 1) { 
                  blah[2]='\0';
                  if (sscanf(blah, "%02d", &hour) == 1) {
                    handle->hours = hour;
                    handle->mins  = min;
                    handle->secs  = sec;
                    handle->statusflags |= GPS_TIME;
                  }
                }
              }
            }
            cur=next;
          }
          break;
      }

      /* reset */ 
      handle->bufpos = 0;   
      handle->packfinished = 0;
      packs++;
    }
  }

  /* If we receive any valid packets, we've got a good  */
  /* link to the GPS and it is operating comm-wise.     */
  /* This can be used to distinguish between situations */
  /* where the GPS has no position fix, is reporting    */
  /* more slowly than expected, or is just plain dead.  */
  if (packs > 0)
    handle->statusflags |= GPS_COMM;

  handle->serialno += packs;

  return packs;
}


int gps_get_status(GPSHandle voidhandle) {
  gpshandle * handle = (gpshandle *) voidhandle;
  int events;

  if (handle == NULL)
    return -1;
   
  events = gps_update(handle); /* check for new data */

  return events; 
}

int gps_get_fixquality(GPSHandle voidhandle, int *fixq) {
  gpshandle * handle = (gpshandle *) voidhandle;
  if (handle->statusflags & GPS_FIXQ) {
    *fixq = handle->fixquality;
    return 0;
  }
  return 1;
}

int gps_get_numsats(GPSHandle voidhandle, int *sats) {
  gpshandle * handle = (gpshandle *) voidhandle;
  if (handle->statusflags & GPS_SATS) {
    *sats = handle->numsats;
    return 0;
  }
  return 1;
}

int gps_get_dop(GPSHandle voidhandle, float *dop) {
  gpshandle * handle = (gpshandle *) voidhandle;
  if (handle->statusflags & GPS_DOP) {
    *dop = handle->dop;
    return 0;
  }
  return 1;
}

int gps_get_date_utc(GPSHandle voidhandle, int *day, int *month, int *year) {
  gpshandle * handle = (gpshandle *) voidhandle;
  if (handle->statusflags & GPS_DATE) {
    *day = handle->day;
    *month = handle->month;
    *year = handle->year;   
    return 0;
  }
  return 1;
}

int gps_get_time_utc(GPSHandle voidhandle, int *hour, int *mins, int *secs) {
  gpshandle * handle = (gpshandle *) voidhandle;
  if (handle->statusflags & GPS_TIME) {
    *hour = handle->hours;
    *mins = handle->mins;
    *secs = handle->secs;   
    return 0;
  }
  return 1;
}

int gps_sprint_date_utc(GPSHandle voidhandle, char *buf) {
  gpshandle * handle = (gpshandle *) voidhandle;
  if (handle->statusflags & GPS_DATE) {
    sprintf(buf, "%02d/%02d/%04d", handle->month, handle->day, handle->year);   
    return 0;
  }
  return 1;
}

int gps_sprint_time_utc(GPSHandle voidhandle, char *buf) {
  gpshandle * handle = (gpshandle *) voidhandle;
  if (handle->statusflags & GPS_TIME) {
    sprintf(buf, "%02d:%02d:%02d", handle->hours, handle->mins, handle->secs);   
    return 0;
  }
  return 1;
}

int gps_systime_diff(GPSHandle voidhandle, int *diffsecs) {
  gpshandle * handle = (gpshandle *) voidhandle;
  if ((handle->statusflags & GPS_TIME) && (handle->statusflags & GPS_DATE)) {
    int diff = 0;
    int month, day, year, hour, mins, secs;
    time_t systp, utctp, gpstp;
    struct tm *utc;
    struct tm gpstime;

    if (gps_get_date_utc(voidhandle, &day, &month, &year)) {
      return 1;
    }
    if (gps_get_time_utc(voidhandle, &hour, &mins, &secs)) {
      return 1;
    }
    time(&systp);
    utc = gmtime(&systp); /* get UTC time from system */
    if (!utc) {
      return 1;
    }

    gpstime.tm_sec   = secs;
    gpstime.tm_min   = mins;
    gpstime.tm_hour  = hour;
    gpstime.tm_mday  = day;
    gpstime.tm_mon   = month - 1;
    gpstime.tm_year  = year - 1900;
    gpstime.tm_isdst = 0;
    gpstp = mktime(&gpstime);

    utc->tm_isdst = 0;
    utctp = mktime(utc);
 
    diff = (int) difftime(utctp, gpstp); 

    *diffsecs = diff; /* difference in seconds */
 
    return 0;
  }  
  return 1;
}

int gps_get_lat_decimal(GPSHandle voidhandle, float *lat) {
  gpshandle * handle = (gpshandle *) voidhandle;
  if (handle->statusflags & GPS_FIXLAT) {
    if (handle->latdir == 'N') 
      *lat = handle->latitude;
    else   
      *lat = -handle->latitude;
    return 0;
  }
  return 1;
}

int gps_get_long_decimal(GPSHandle voidhandle, float *lon) {
  gpshandle * handle = (gpshandle *) voidhandle;
  if (handle->statusflags & GPS_FIXLONG) {
    if (handle->latdir == 'E') 
      *lon = handle->longtitude;
    else   
      *lon = -handle->longtitude;
    return 0;
  }
  return 1;
}

int gps_sprint_lat(GPSHandle voidhandle, char *buf) {
  gpshandle * handle = (gpshandle *) voidhandle;
  if (handle->statusflags & GPS_FIXLAT) {
    sprintf(buf, "%5.2f %c", handle->latitude, handle->latdir);
    return 0;
  }
  return 1;
}


int gps_sprint_long(GPSHandle voidhandle, char *buf) {
  gpshandle * handle = (gpshandle *) voidhandle;
  if (handle->statusflags & GPS_FIXLONG) {
    sprintf(buf, "%5.2f %c", handle->longtitude, handle->longdir);
    return 0;
  }
  return 1;
}

