/*
 * util.c - utility routines for platform independent timers, etc.
 *
 * $Id: util.c,v 1.1 2007/09/02 06:42:04 johns Exp $
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

#include "util.h"

/* 
 * platform specific routines
 */

#if defined(_MSC_VER)
#include <windows.h>

typedef struct {
  DWORD starttime;
  DWORD endtime;
} rt_timer;

void rt_timer_start(rt_timerhandle v) {
  rt_timer * t = (rt_timer *) v;
  t->starttime = GetTickCount();
}

void rt_timer_stop(rt_timerhandle v) {
  rt_timer * t = (rt_timer *) v;
  t->endtime = GetTickCount();
}

float rt_timer_time(rt_timerhandle v) {
  rt_timer * t = (rt_timer *) v;
  float ttime;

  ttime = (t->endtime - t->starttime) / 1000.0f;

  return ttime;
}

#elif 1

#include <unistd.h>
#include <sys/time.h>

/* if we're on a Unix with gettimeofday() we'll use newer timers */
typedef struct {
  struct timeval starttime, endtime;
#ifndef VMS
  struct timezone tz;
#endif
} rt_timer;

void rt_timer_start(rt_timerhandle v) {
  rt_timer * t = (rt_timer *) v;
#ifdef VMS
  gettimeofday(&t->starttime, NULL);
#else
  gettimeofday(&t->starttime, &t->tz);
#endif
}

void rt_timer_stop(rt_timerhandle v) {
  rt_timer * t = (rt_timer *) v;
#ifdef VMS
  gettimeofday(&t->endtime, NULL);
#else
  gettimeofday(&t->endtime, &t->tz);
#endif
}

float rt_timer_time(rt_timerhandle v) {
  rt_timer * t = (rt_timer *) v;
  double ttime, start, end;

  start = (t->starttime.tv_sec + 1.0 * t->starttime.tv_usec / 1000000.0);
    end = (t->endtime.tv_sec + 1.0 * t->endtime.tv_usec / 1000000.0);
  ttime = end - start;

  return (float) ttime;
}

#else

#include <time.h>

typedef struct {
  struct timespec starttime;
  struct timespec endtime;
} rt_timer;

void rt_timer_start(rt_timerhandle v) {
  rt_timer * t = (rt_timer *) v;
  clock_gettime(CLOCK_REALTIME, &t->starttime);
}

void rt_timer_stop(rt_timerhandle v) {
  rt_timer * t = (rt_timer *) v;
  clock_gettime(CLOCK_REALTIME, &t->endtime);
}

float rt_timer_time(rt_timerhandle v) {
  rt_timer * t = (rt_timer *) v;
  float ttime, start, end;

  start = (t->starttime.tv_sec + 1.0 * t->starttime.tv_nsec / 1000000000.0);
    end = (t->endtime.tv_sec + 1.0 * t->endtime.tv_nsec / 1000000000.0);
  ttime = end - start;

  return (float) ttime;
}
#endif

/*
 * system independent routines to create and destroy timers 
 */ 
rt_timerhandle rt_timer_create(void) {
  rt_timer * t;
  t = (rt_timer *) malloc(sizeof(rt_timer));
  memset(t, 0, sizeof(rt_timer));
  return t;
}

void rt_timer_destroy(rt_timerhandle v) {
  free(v);
}

float rt_timer_timenow(rt_timerhandle v) {
  rt_timer_stop(v);
  return rt_timer_time(v);
}


/*
 * message timer stuff
 */

/* get the time of day from the system clock, and store it (in seconds) */
double time_of_day(void) {
#if defined(_MSC_VER)
  double t;

  t = GetTickCount();
  t = t / 1000.0;

  return t;
#else
  struct timeval tm;
  struct timezone tz;

  gettimeofday(&tm, &tz);
  return((double)(tm.tv_sec) + (double)(tm.tv_usec)/1000000.0);
#endif
}

/* initialize status message timer */
void msg_timer_init(msgtimer *t, double updatetime) {
  if (t != NULL) {
    t->lasttime = time_of_day();
    t->updatetime = updatetime;
  }
}

/* return true if it's time to print a status update message */
int msg_timer_timeout(msgtimer *t) {
  double curtime = time_of_day();
  double elapsed = curtime - t->lasttime;
  if (elapsed > t->updatetime) {
    /* reset the clock and return true that our timer expired */
    t->lasttime = time_of_day();
    return 1;
  } else if (elapsed < 0) {
    /* time went backwards, best reset our clock! */
    t->lasttime = time_of_day();
  }
  return 0;
}


void rt_sleep(int msec) {
#if defined(_MSC_VER) || defined(WIN32)
  Sleep(msec);
#else
  usleep(msec * 1000);
#endif
}

int countdown(const char *msgformat, float seconds) {
  int done;
  rt_timerhandle intervaltimer;
  intervaltimer = rt_timer_create();
  rt_timer_start(intervaltimer);

  /* spin waiting on time interval... */
  done = 0;
  while (!done) {
    float curtime, remaining;
        rt_timer_stop(intervaltimer);
    curtime = rt_timer_time(intervaltimer);
    remaining = seconds - curtime;
    if (remaining > 3) {
      printf(msgformat, remaining);
      fflush(stdout);
      rt_sleep(1000);
    } else if (remaining > 0.1) {
      printf(msgformat, remaining);
      fflush(stdout);
    } else if (remaining <= 0) {
      done = 1;
    }
  }
  printf(msgformat, 0);
  printf("\n");

  rt_timer_destroy(intervaltimer);
  return 0;
}


/* Sets the current date and time in format 'yyyy-mm-ddThh:mm:ss'. */
/* returns zero if UTC, 1 if local time                            */
int get_utc_datetime(char *timestr) {
  time_t tp;
  struct tm *utc;
  struct tm *local;

  time(&tp);         /* get current time */
  utc = gmtime(&tp); /* get UTC time     */
  if (utc) {
    /* FITS friendly time format string */
    strftime(timestr, 25, "%Y-%m-%dT%H:%M:%S", utc);
    return 0;
  } 

  /* Failed to get UTC time, so return local time and flag return value */
  local = localtime(&tp);
  strftime(timestr, 25, "%Y-%m-%dT%H:%M:%S", local);
  return 1;
}


/* Sets the current date and time in format 'hh:mm:ss'. */
/* returns zero if UTC, 1 if local time                            */
int get_utc_time(char *timestr) {
  time_t tp;
  struct tm *utc;
  struct tm *local;

  time(&tp);         /* get current time */
  utc = gmtime(&tp); /* get UTC time     */
  if (utc) {
    /* FITS friendly time format string */
    strftime(timestr, 25, "%H:%M:%S", utc);
    return 0;
  } 

  /* Failed to get UTC time, so return local time and flag return value */
  local = localtime(&tp);
  strftime(timestr, 25, "%H:%M:%S", local);
  return 1;
}


#if 0
/* report fractional secons value for the system clock */
int get_utc_secfrac(float *frac) {
  struct timeval t;
#ifndef VMS
  struct timezone tz;
#endif

#ifdef VMS
  gettimeofday(&t, NULL);
#else
  gettimeofday(&t, &tz);
#endif

 *frac = (double)(t.tv_usec)/1000000.0;

  return 0; /* all is well */
}
#endif

// do upper-case comparison, up to n characters
int strupncmp(const char *a, const char *b, int n) {
#if defined(_MSC_VER)
   while (n-- > 0) {
      if (toupper(*a) != toupper(*b)) {
         return toupper(*b) - toupper(*a);
      }
      if (*a == 0) return 0;
      a++; b++;
   }
   return 0;
#else
   return strncasecmp(a, b, n);
#endif
}


